from __future__ import annotations

from collections.abc import Callable

from django.http import HttpRequest

import time

import requests
from sentry.auth.provider import MigratingIdentityId
from sentry.auth.providers.oauth2 import OAuth2Callback, OAuth2Login, OAuth2Provider
from sentry.auth.services.auth.model import RpcAuthProvider
from sentry.organizations.services.organization.model import RpcOrganization
from sentry.plugins.base.response import DeferredResponse

from .constants import (
    AUTHORIZATION_ENDPOINT,
    CLIENT_ID,
    CLIENT_SECRET,
    DATA_VERSION,
    ISSUER,
    SCOPE,
    TOKEN_ENDPOINT,
    USERINFO_ENDPOINT,
)
from .views import FetchUser, oidc_configure_view


class OIDCLogin(OAuth2Login):
    authorize_url = AUTHORIZATION_ENDPOINT
    client_id = CLIENT_ID
    scope = SCOPE

    def __init__(self, client_id, domains=None):
        self.domains = domains
        super().__init__(client_id=client_id)

    def get_authorize_params(self, state, redirect_uri):
        params = super().get_authorize_params(state, redirect_uri)
        # TODO(dcramer): ideally we could look at the current resulting state
        # when an existing auth happens, and if they're missing a refresh_token
        # we should re-prompt them a second time with ``approval_prompt=force``
        params["approval_prompt"] = "force"
        params["access_type"] = "offline"
        return params


class OIDCProvider(OAuth2Provider):
    name = ISSUER

    def __init__(self, domain=None, domains=None, version=None, **config):
        if domain:
            if domains:
                domains.append(domain)
            else:
                domains = [domain]
        self.domains = domains
        # if a domain is not configured this is part of the setup pipeline
        # this is a bit complex in Sentry's SSO implementation as we don't
        # provide a great way to get initial state for new setup pipelines
        # vs missing state in case of migrations.
        if domains is None:
            version = DATA_VERSION
        else:
            version = None
        self.version = version
        super().__init__(**config)

    def get_client_id(self):
        return CLIENT_ID

    def get_client_secret(self):
        return CLIENT_SECRET

    def get_configure_view(
        self,
    ) -> Callable[[HttpRequest, RpcOrganization, RpcAuthProvider], DeferredResponse]:
        return oidc_configure_view

    def get_auth_pipeline(self):
        return [
            OIDCLogin(domains=self.domains, client_id=self.get_client_id()),
            OAuth2Callback(
                access_token_url=TOKEN_ENDPOINT,
                client_id=self.get_client_id(),
                client_secret=self.get_client_secret(),
            ),
            FetchUser(domains=self.domains, version=self.version),
        ]

    def get_refresh_token_url(self):
        return TOKEN_ENDPOINT

    def build_config(self, state):
        return {"domains": [state["domain"]], "version": DATA_VERSION}

    def get_user_info(self, bearer_token):
        endpoint = USERINFO_ENDPOINT
        bearer_auth = "Bearer " + bearer_token
        retry_codes = [429, 500, 502, 503, 504]
        for retry in range(10):
            if 10 < retry:
                return {}
            r = requests.get(
                endpoint + "?schema=openid",
                headers={"Authorization": bearer_auth},
                timeout=20.0,
            )
            if r.status_code in retry_codes:
                wait_time = 2**retry * 0.1
                time.sleep(wait_time)
                continue
            return r.json()

    def build_identity(self, state):
        data = state["data"]
        user_data = state["user"]

        bearer_token = data["access_token"]
        user_info = self.get_user_info(bearer_token)

        # XXX(epurkhiser): We initially were using the email as the id key.
        # This caused account dupes on domain changes. Migrate to the
        # account-unique sub key.
        # user_id = MigratingIdentityId(id=user_data["sub"], legacy_id=user_data["email"])
        user_email = user_info.get("email")
        email = user_email if user_email else (user_data["sub"] + '@corp.netease.com')
        user_id = MigratingIdentityId(id=user_data["sub"], legacy_id=email)

        return {
            "id": user_id,
            "email": email,
            "name": user_data["sub"],
            "data": self.get_oauth_data(data),
            "email_verified": user_info.get("email_verified"),
        }



import requests
from django.conf import settings

AUTHORIZATION_ENDPOINT = getattr(settings, "OIDC_AUTHORIZATION_ENDPOINT", None)
TOKEN_ENDPOINT = getattr(settings, "OIDC_TOKEN_ENDPOINT", None)
CLIENT_ID = getattr(settings, "OIDC_CLIENT_ID", None)
CLIENT_SECRET = getattr(settings, "OIDC_CLIENT_SECRET", None)
USERINFO_ENDPOINT = getattr(settings, "OIDC_USERINFO_ENDPOINT", None)
SCOPE = getattr(settings, "OIDC_SCOPE", "openid email")
WELL_KNOWN_SCHEME = "/.well-known/openid-configuration"
ERR_INVALID_RESPONSE = (
    "Unable to fetch user information from provider.  Please check the log."
)
ISSUER = None

DATA_VERSION = "1"

OIDC_DOMAIN = getattr(settings, "OIDC_DOMAIN", None)
if OIDC_DOMAIN:
    WELL_KNOWN_URL = OIDC_DOMAIN.strip("/") + WELL_KNOWN_SCHEME
    well_known_values = requests.get(WELL_KNOWN_URL, timeout=2.0).json()
    if well_known_values:
        USERINFO_ENDPOINT = well_known_values["userinfo_endpoint"]
        AUTHORIZATION_ENDPOINT = well_known_values["authorization_endpoint"]
        TOKEN_ENDPOINT = well_known_values["token_endpoint"]
        ISSUER = well_known_values["issuer"]


config_issuer = getattr(settings, "OIDC_ISSUER", None)
if config_issuer:
    ISSUER = config_issuer



from __future__ import annotations

import logging

from django.http import HttpRequest
from rest_framework.response import Response

from sentry.auth.services.auth.model import RpcAuthProvider
from sentry.auth.view import AuthView
from sentry.utils import json
from sentry.organizations.services.organization.model import RpcOrganization
from sentry.plugins.base.response import DeferredResponse
from sentry.utils.signing import urlsafe_b64decode

from .constants import ERR_INVALID_RESPONSE, ISSUER

logger = logging.getLogger("sentry.auth.oidc")


class FetchUser(AuthView):
    def __init__(self, domains, version, *args, **kwargs):
        self.domains = domains
        self.version = version
        super().__init__(*args, **kwargs)

    def dispatch(self, request: HttpRequest, helper) -> Response:  # type: ignore
        data = helper.fetch_state("data")

        try:
            id_token = data["id_token"]
        except KeyError:
            logger.error("Missing id_token in OAuth response: %s" % data)
            return helper.error(ERR_INVALID_RESPONSE)

        try:
            _, payload, _ = map(urlsafe_b64decode, id_token.split(".", 2))
        except Exception as exc:
            logger.error("Unable to decode id_token: %s" % exc, exc_info=True)
            return helper.error(ERR_INVALID_RESPONSE)

        try:
            payload = json.loads(payload)
        except Exception as exc:
            logger.error("Unable to decode id_token payload: %s" % exc, exc_info=True)
            return helper.error(ERR_INVALID_RESPONSE)

        # if not payload.get("email"):
        #     logger.error("Missing email in id_token payload: %s" % id_token)
        #     return helper.error(ERR_INVALID_RESPONSE)

        # support legacy style domains with pure domain regexp
        if self.version is None:
            # domain = extract_domain(payload["email"])
            domain = "corp.netease.com"
            payload["email"] = payload["sub"] + "@" + domain
        else:
            domain = payload.get("hd")

        helper.bind_state("domain", domain)
        helper.bind_state("user", payload)

        return helper.next_step()


def oidc_configure_view(
    request: HttpRequest, organization: RpcOrganization, auth_provider: RpcAuthProvider
) -> DeferredResponse:
    config = auth_provider.config
    if config.get("domain"):
        domains: list[str] | None
        domains = [config["domain"]]
    else:
        domains = config.get("domains")
    return DeferredResponse(
        "oidc/configure.html", {"provider_name": ISSUER or "", "domains": domains or []}
    )


def extract_domain(email):
    return email.rsplit("@", 1)[-1]



import pytest
from sentry import auth
from sentry.auth.exceptions import IdentityNotValid
from sentry.models import AuthIdentity, AuthProvider
from sentry.testutils.cases import TestCase
from sentry.testutils.silo import control_silo_test

from oidc.constants import DATA_VERSION
from oidc.provider import OIDCProvider


@control_silo_test
class OIDCProviderTest(TestCase):
    def setUp(self):
        self.auth_provider_inst = AuthProvider.objects.create(
            provider="oidc", organization_id=self.organization.id
        )
        auth.register("oidc", OIDCProvider)
        super().setUp()

    def test_refresh_identity_without_refresh_token(self):
        auth_identity = AuthIdentity.objects.create(
            auth_provider=self.auth_provider_inst,
            user=self.user,
            data={"access_token": "access_token"},
        )

        provider = self.auth_provider_inst.get_provider()

        with pytest.raises(IdentityNotValid):
            provider.refresh_identity(auth_identity)

    def test_handles_multiple_domains(self):
        self.auth_provider_inst.update(config={"domains": ["example.com"]})

        provider = self.auth_provider_inst.get_provider()
        assert provider.domains == ["example.com"]

    def test_handles_legacy_single_domain(self):
        self.auth_provider_inst.update(config={"domain": "example.com"})

        provider = self.auth_provider_inst.get_provider()
        assert provider.domains == ["example.com"]

    def test_build_config(self):
        provider = self.auth_provider_inst.get_provider()
        state = {
            "domain": "example.com",
            "user": {
                "iss": "accounts.google.com",
                "at_hash": "HK6E_P6Dh8Y93mRNtsDB1Q",
                "email_verified": "true",
                "sub": "10769150350006150715113082367",
                "azp": "1234987819200.apps.googleusercontent.com",
                "email": "jsmith@example.com",
                "aud": "1234987819200.apps.googleusercontent.com",
                "iat": 1353601026,
                "exp": 1353604926,
                "hd": "example.com",
            },
        }
        result = provider.build_config(state)
        assert result == {"domains": ["example.com"], "version": DATA_VERSION}



from django.apps import AppConfig


class Config(AppConfig):
    name = "oidc"

    def ready(self):
        from sentry import auth

        from .provider import OIDCProvider

        auth.register("oidc", OIDCProvider)



#!/usr/bin/env python
from __future__ import absolute_import
from setuptools import setup, find_packages

# with open('README.rst') as readme_file:
#     readme = readme_file.read()

setup(
    name='kkauth-oidc',
    version='0.0.10',
    author='yliiii',
    author_email='',
    url='',
    description='OpenID Connect authentication provider for Sentry',
    long_description='',
    license='Apache 2.0',
    package_data={'': ['**/*.html']},
    packages=find_packages(exclude=['tests']),
    zip_safe=False,
    include_package_data=True,
    entry_points={
        'sentry.apps': [
            'oidc = oidc.apps.Config',
        ],
    },
    python_requires='>=3.11, <4.0',
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: Other/Proprietary License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Topic :: Software Development',
    ],
    long_description_content_type='text/x-rst',
)
